from .utils import join_openai_response

import asyncio
import logging
import openai


logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO,
                    format='[%(asctime)s] [%(levelname)s] - %(message)s')

class CodeAnalyzer:
    def __init__(self, api_key:str, model_engine:str="gpt-3.5-turbo"):
        if not isinstance(api_key, str):
            raise ValueError('API key should be str')
        
        openai.api_key = api_key
        self.model_engine = model_engine

    async def find_bugs_in_code(self, code):
        prompt = f"What are the bugs in this code? Also find logical errors or issues with implementation\n\n{code}\n\nBugs:"
        messages = [{"role":"user", "content":prompt}]

        response = openai.ChatCompletion.create(
            model=self.model_engine,
            messages=messages,
        )

        return response
    
    async def find_vulns_in_code(self, code):
        prompt = f"What are the vulnerabilities in this code? and how can we fix it?\n\n{code}"
        messages = [{"role":"user", "content":prompt}]

        response = openai.ChatCompletion.create(
            model=self.model_engine,
            messages=messages,
        )

        return response
    
    async def analyze_code(self, file_name:str, code:str):
        logger.info(f'Analyzing code of file: {file_name}')

        tasks = [
            # for bug
            asyncio.ensure_future(self.find_bugs_in_code(code)),
            # for vulns
            asyncio.ensure_future(self.find_vulns_in_code(code)),
        ]
        responses = await asyncio.gather(*tasks)

        bugs_response = responses[0]
        vulns_response = responses[1]

        # extract data 
        if bugs_response:
            bugs_response = join_openai_response(bugs_response)
        else:
            logging.warning(f'unable to join messsages for {file_name} bug response')

        if vulns_response:
            vulns_response = join_openai_response(vulns_response)
        else:
            logging.warning(f'unable to join messsages for {file_name} vulns response')

        analyzed_data = {
            "file_name": file_name,
            "bugs_response": bugs_response,
            "vulns_response": vulns_response,
            "code":code,
        }

        return analyzed_data

    async def analyze_git_changes(self, diff_changes:dict):
        '''
        accepts data in dict form from GitHandler component
        '''
        tasks = []
        for diff in diff_changes:
            file_name = diff.get('file_path')
            changed_lines = diff.get('changed_lines')

            tasks.append(asyncio.ensure_future(self.analyze_code(file_name=file_name, code=changed_lines)))
        
        analyzed_code_snippets = await asyncio.gather(*tasks)

        return analyzed_code_snippets


from git.repo import Repo
from os.path import isdir

import logging


logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO,
                    format='[%(asctime)s] [%(levelname)s] - %(message)s')


class GitDirectoryNotFound(Exception):
    pass

class GitHandler:
    def __init__(self, repo_dir:str) -> None:
        '''
        initialize git repo directory for getting project information
        '''

        if not isdir(repo_dir):
            msg = f"{repo_dir} directory not found."
            logger.exception(msg)
            raise GitDirectoryNotFound(msg)
        
        self._repo_dir = repo_dir
        self._repo = Repo(repo_dir)


    def __get_commit_hashes(self, commit_no:int=1):
        '''
        get commit hashes
        '''
        commits = list(self._repo.iter_commits('HEAD'))
        logger.info(f'Total commits: {len(commits)}')

        if len(commits) > 1 and len(commits) >= commit_no:
            latest_commit_hash = commits[0].hexsha
            target_commit_hash = commits[commit_no].hexsha
            logger.info(f'Found {commit_no} commits from HEAD')
            return latest_commit_hash, target_commit_hash
        
        elif len(commits) > 1:
            latest_commit_hash = commits[0].hexsha
            target_commit_hash = commits[1].hexsha
            logger.info(f'couldn\'t find {commit_no} commits from HEAD. using previous commit hash as target commit.')
            return latest_commit_hash, target_commit_hash

        logger.warning('No previous commit found.')

        # return HEAD and empty tree hash 
        return 'HEAD', '4b825dc642cb6eb9a060e54bf8d69288fbee4904'



    def get_diff(self, commit_no:int=None):
        '''
        get git difference between previous.
        param: commits_no (int), no of commits to be analyzed from current head.
        '''
        diffs = []

        latest_commit_hash, target_commit_hash = self.__get_commit_hashes(commit_no)

        diff_str = self._repo.git.diff('--no-renames', '-U0', target_commit_hash, latest_commit_hash)
        
        # Split the diff string into lines
        diff_lines = diff_str.split('\n')


        # Initialize variables for the current file path and changed lines
        file_path = None
        changed_lines = []

        # Iterate over the lines in the diff and print the changed ones
        for line in diff_lines:
            if line.startswith('diff --git'):
                # If this is the start of a new file diff, print the changed lines for the previous file (if any)
                if file_path is not None:
                    # pass
                    # print('---', file_path, '---')
                    # for changed_line in changed_lines:
                        # print(changed_line)
                
                    diffs.append({
                        "file_path": file_path,
                        "changed_lines": changed_lines
                    })

                # Initialize variables for the new file diff
                file_path = line.split(' b/')[-1]
                changed_lines = []

            elif line.startswith('++'):
                continue

            elif line.startswith('+'):
                # If this is a changed line, add it to the list of changed lines for the current file
                changed_lines.append(line.strip().removeprefix('+'))
        
        # Print the changed lines for the last file
        if file_path is not None:
             diffs.append({
                "file_path": file_path,
                "changed_lines": changed_lines
            })
             
            # print('---', file_path, '---')
            # for changed_line in changed_lines:
                # print(changed_line)

        return diffs


from argparse import ArgumentParser
from asyncio import run
from os import environ


from .chatgpt_api import CodeAnalyzer 
from .git_handler import GitHandler
from .utils import join_diffs, json_to_file, analyzed_data_to_console


OPEN_API_KEY = environ.get('OPEN_API_KEY', False)
code_analyzer = CodeAnalyzer(
    api_key=OPEN_API_KEY,
    model_engine='gpt-3.5-turbo',
)


if __name__ == '__main__':
    parser = ArgumentParser(prog='shift_left_secure')
    parser.add_argument('-d', '--directory', help='directory of git project', dest='directory', type=str, required=True)
    parser.add_argument('-c', '--commit_hash', help='no of commits to be analyzed from current HEAD', dest='commits', type=int, default=1)
    parser.add_argument('-o', '--output', help='output path to json file', dest='output_path' ,required=False, default=False, type=str)

    args = parser.parse_args()

    # get diff changes
    git = GitHandler(repo_dir=args.directory)
    diff_changes = join_diffs(git.get_diff(args.commits))

    analyzed_code_snippets = run(code_analyzer.analyze_git_changes(diff_changes))

    analyzed_data_to_console(analyzed_code_snippets)

    if args.output_path:
        json_to_file(args.output_path,analyzed_code_snippets)



from json import dumps
import logging


logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO,
                    format='[%(asctime)s] [%(levelname)s] - %(message)s')

def join_diffs(diffs:list):
    for file_change_data in diffs:
        changed_lines = '\n'.join(file_change_data.get('changed_lines',[]))
        file_change_data['changed_lines'] = changed_lines

    return diffs

def join_openai_response(response:dict):
    choices = response.get('choices')
    if not choices:
        return False
    
    message = ''
    for choice in choices:
        message += choice.get('message', {'content':''}).get('content','')

    return message

def json_to_file(file_path:str, data:dict or list):
    '''
    dumps json data to a file
    '''
    if not file_path.endswith('.json'):
        file_path += '.json'
    
    try:
        with open(file_path, 'w') as f:
            f.write(dumps(data))
            logger.info(f'responses stored in file: {file_path}')

    except Exception as e:
        logger.exception(e)

    return True

def analyzed_data_to_console(analyzed_data:list):
    if not isinstance(analyzed_data, list):
        logger.warning('analyzed_data should be a list')
        return 

    for code_analysis in analyzed_data:
        file_name = code_analysis.get('file_name')
        bugs_response = code_analysis.get('bugs_response')
        vulns_response = code_analysis.get('vulns_response')

        print('FILE NAME:\t', file_name,'\n')
        print('BUGS:\n',bugs_response, '\n')
        print('VULNS:\n', vulns_response, '\n')
        print('-'*20)
